/*
 * Copyright (c) 1998, Larry Lile
 * All rights reserved.
 *
 * For latest sources and information on this driver, please
 * go to http://anarchy.stdio.com.
 *
 * Questions, comments or suggestions should be directed to
 * Larry Lile <lile@stdio.com>.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice unmodified, this list of conditions, and the following
 *    disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: if_ol.c,v 1.1 1999/08/28 14:34:53 lile Exp $
 */ 

#include "bpf.h"

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/sockio.h>
#include <sys/mbuf.h>
#include <sys/malloc.h>
#include <sys/kernel.h>
#include <sys/socket.h>

#include <net/if.h>
#include <net/if_arp.h>
#include <net/iso88025.h>
#include <net/ethernet.h>
#include <net/if_dl.h>
#include <net/if_media.h>
#include <net/iso88025.h>

#if NBPF > 0
#include <net/bpf.h>
#endif

#include <vm/vm.h>              /* for vtophys */
#include <vm/pmap.h>            /* for vtophys */
#include <machine/clock.h>      /* for DELAY */
#include <machine/bus_memio.h>
#include <machine/bus_pio.h>
#include <machine/bus.h>
#include <machine/resource.h>
#include <sys/bus.h>
#include <sys/rman.h>

#include <pci/pcireg.h>
#include <pci/pcivar.h>

#include "contrib/dev/ol/trlld.h"

#ifndef DEBUG_MASK
#define DEBUG_MASK 0x0000
#endif

#define DEBUG_POLL	0x0001
#define DEBUG_INT	0x0002

#define PCI_VENDOR_OLICOM 0x108D

#define MIN(A,B) (((A) < (B)) ? (A) : (B))
#define MIN3(A,B,C) (MIN(A, (MIN(B, C))))

char *AdapterName[] = {
	/*  0 */ "Olicom XT Adapter [unsupported]",
	/*  1 */ "Olicom OC-3115",
	/*  2 */ "Olicom ISA 16/4 Adapter (OC-3117)",
	/*  3 */ "Olicom ISA 16/4 Adapter (OC-3118)",
	/*  4 */ "Olicom MCA 16/4 Adapter (OC-3129) [unsupported]",
	/*  5 */ "Olicom MCA 16/4 Adapter (OC-3129) [unsupported]",
	/*  6 */ "Olicom MCA 16/4 Adapter (OC-3129) [unsupported]",
	/*  7 */ "Olicom EISA 16/4 Adapter (OC-3133)",
	/*  8 */ "Olicom EISA 16/4 Adapter (OC-3133)",
	/*  9 */ "Olicom EISA 16/4 Server Adapter (OC-3135)",
	/* 10 */ "Olicom PCI 16/4 Adapter (OC-3136)",
	/* 11 */ "Olicom PCI 16/4 Adapter (OC-3136)",
	/* 12 */ "Olicom PCI/II 16/4 Adapter (OC-3137)",
	/* 13 */ "Olicom PCI 16/4 Adapter (OC-3139)",
	/* 14 */ "Olicom RapidFire 3140 16/4 PCI Adapter (OC-3140)",
	/* 15 */ "Olicom RapidFire 3141 Fiber Adapter (OC-3141)",
	/* 16 */ "Olicom PCMCIA 16/4 Adapter (OC-3220) [unsupported]",
	/* 17 */ "Olicom PCMCIA 16/4 Adapter (OC-3121, OC-3230, OC-3232) [unsupported]",
	/* 18 */ "Olicom PCMCIA 16/4 Adapter (OC-3250)",
	/* 19 */ "Olicom RapidFire 3540 4/16/100 Adapter (OC-3540)"
};

/*
 * Glue function prototypes for PMW kit IO 
 */

#ifndef TRlldInlineIO
static void DriverOutByte		__P((unsigned short, unsigned char));
static void DriverOutWord		__P((unsigned short, unsigned short));
static void DriverOutDword		__P((unsigned short, unsigned long));
static void DriverRepOutByte		__P((unsigned short, unsigned char  *, int));
static void DriverRepOutWord		__P((unsigned short, unsigned short *, int));
static void DriverRepOutDword		__P((unsigned short, unsigned long  *, int));
static unsigned char  DriverInByte	__P((unsigned short));
static unsigned short DriverInWord	__P((unsigned short));
static unsigned long  DriverInDword	__P((unsigned short));
static void DriverRepInByte		__P((unsigned short, unsigned char  *, int));
static void DriverRepInWord		__P((unsigned short, unsigned short *, int));
static void DriverRepInDword		__P((unsigned short, unsigned long  *, int));
#endif /*TRlldInlineIO*/
static void DriverSuspend		__P((unsigned short));
static void DriverStatus		__P((void *, TRlldStatus_t *));
static void DriverCloseCompleted	__P((void *));
static void DriverStatistics		__P((void *, TRlldStatistics_t *));
static void DriverTransmitFrameCompleted __P((void *, void *, int));
static void DriverReceiveFrameCompleted	__P((void *, int, int, void *, int));

static TRlldDriver_t LldDriver = {
	TRLLD_VERSION,
#ifndef TRlldInlineIO
	DriverOutByte,
	DriverOutWord,
	DriverOutDword,
	DriverRepOutByte,
	DriverRepOutWord,
	DriverRepOutDword,
	DriverInByte,
	DriverInWord,
	DriverInDword,
	DriverRepInByte,
	DriverRepInWord,
	DriverRepInDword,
#endif /*TRlldInlineIO*/
	DriverSuspend,
	DriverStatus,
	DriverCloseCompleted,
	DriverStatistics,
	DriverTransmitFrameCompleted,
	DriverReceiveFrameCompleted,
};

struct ol_rx_buf {
	int			index;
	char			*data;
	u_long			address;
};

struct ol_tx_buf {
	int			index;
	char 			*data;
	u_long			address;
};

#define RING_BUFFER_LEN		16
#define RING_BUFFER(x)		((RING_BUFFER_LEN - 1) & x)
#define RX_BUFFER_LEN		2048
#define TX_BUFFER_LEN		512

struct ol_softc {
	struct arpcom		arpcom;
/*	struct if_media		ifmedia; */
	bus_space_handle_t	ol_bhandle;
	bus_space_tag_t		ol_btag;
	void			*ol_intrhand;
	struct resource		*ol_irq;
	struct resource		*ol_res;
	int			unit;
	int 			state;
#define OL_UNKNOWN	0
#define OL_INIT		1
#define OL_READY	2
#define OL_CLOSING	3
#define OL_CLOSED	4
#define OL_OPENING	5
#define OL_OPEN		6
#define OL_PROMISC	7
	struct ol_rx_buf	rx_ring[RING_BUFFER_LEN];
	int			tx_head, tx_avail, tx_frame;
	struct ol_tx_buf	tx_ring[RING_BUFFER_LEN];
	TRlldTransmit_t		frame_ring[RING_BUFFER_LEN];
	struct mbuf		*restart;
	TRlldAdapter_t		TRlldAdapter;
	TRlldStatistics_t	statistics;
	TRlldStatistics_t	current;
	u_short			AdapterMode;
	u_long			GroupAddress;
	u_long			FunctionalAddress;
	struct callout_handle	ol_poll_ch;
	/*struct callout_handle	ol_stat_ch;*/
};

#define SELF_TEST_POLLS	100

static int ol_pci_probe		__P((device_t));
static int ol_pci_attach	__P((device_t));
static int ol_pci_detach	__P((device_t));
static void ol_pci_shutdown	__P((device_t));
static void ol_intr		__P((void *));

void ol_poll 			__P((void *));
/*void ol_stat 			__P((void *));*/

static void ol_start		__P((struct ifnet *));
static void ol_stop		__P((struct ol_softc *));
static void ol_init		__P((void *));
static int ol_ioctl		__P((struct ifnet *, u_long, caddr_t));


static device_method_t ol_methods[] = {
	DEVMETHOD(device_probe,		ol_pci_probe),
	DEVMETHOD(device_attach,	ol_pci_attach),
	DEVMETHOD(device_detach,	ol_pci_detach),
	DEVMETHOD(device_shutdown,	ol_pci_shutdown),
        { 0, 0 }
};

static driver_t ol_driver = {
	"ol",
	ol_methods,
	sizeof(struct ol_softc)
};

static devclass_t ol_devclass;

DRIVER_MODULE(ol, pci, ol_driver, ol_devclass, 0, 0);

static int ol_pci_probe(dev)
	device_t	dev;
{
	int			i, rc;
	char 			PCIConfigHeader[64];
	TRlldAdapterConfig_t	config;
	

	if ((pci_get_vendor(dev) == PCI_VENDOR_OLICOM) && 
           ((pci_get_device(dev) == 0x0001) || (pci_get_device(dev) == 0x0004) ||
            (pci_get_device(dev) == 0x0005) || (pci_get_device(dev) == 0x0007) ||
            (pci_get_device(dev) == 0x0008))) {

		for (i = 0; i < sizeof(PCIConfigHeader); i++)
			PCIConfigHeader[i] = pci_read_config(dev, i, 1);

		rc = TRlldPCIConfig(&LldDriver, &config, PCIConfigHeader);
		if (rc == TRLLD_PCICONFIG_FAIL) {
			device_printf(dev, "TRlldPciConfig failed!\n");
			return(ENXIO);
		}
		if (rc == TRLLD_PCICONFIG_VERSION) {
			device_printf(dev, "wrong LLD version\n");
			return(ENXIO);
		}
		device_set_desc(dev, AdapterName[config.type]);
		return(0);
	}
	return(ENXIO);
}
static int ol_pci_attach(dev)
	device_t	dev;
{
        int 			poll = 0, i, s, rc = 0, rid, scratch_size;
	u_long 			command;
	char 			PCIConfigHeader[64];
	struct ol_softc		*sc = device_get_softc(dev);
	struct ifnet		*ifp = &sc->arpcom.ac_if;
	TRlldAdapterConfig_t	config;

        s = splimp();

       	bzero(sc, sizeof(struct ol_softc));
	sc->unit = device_get_unit(dev);
	sc->state = OL_UNKNOWN;

	for (i = 0; i < sizeof(PCIConfigHeader); i++)
		PCIConfigHeader[i] = pci_read_config(dev, i, 1);

	switch(TRlldPCIConfig(&LldDriver, &config, PCIConfigHeader)) {
		case TRLLD_PCICONFIG_OK:
			break;
		case TRLLD_PCICONFIG_SET_COMMAND:
			device_printf(dev, "enabling bus master mode\n");
			command = pci_read_config(dev, PCI_COMMAND_STATUS_REG, 4);
			pci_write_config(dev, PCI_COMMAND_STATUS_REG, 
				(command | PCIM_CMD_BUSMASTEREN), 4);
			command = pci_read_config(dev, PCI_COMMAND_STATUS_REG, 4);
			if (!(command & PCIM_CMD_BUSMASTEREN)) {
				device_printf(dev, "failed to enable bus master mode\n");
				goto config_failed;
			}
			break;
		case TRLLD_PCICONFIG_FAIL:
			device_printf(dev, "TRlldPciConfig failed!\n");
			goto config_failed;
			break;
		case TRLLD_PCICONFIG_VERSION:
			device_printf(dev, "wrong LLD version\n");
			goto config_failed;
			break;
	}
	device_printf(dev, "MAC address %6D\n", config.macaddress, ":");

	scratch_size = TRlldAdapterSize();
	if (bootverbose)
		device_printf(dev, "adapter memory block size %d bytes\n", scratch_size);
	sc->TRlldAdapter = (TRlldAdapter_t)malloc(scratch_size, M_DEVBUF, M_NOWAIT);
	if (sc->TRlldAdapter == NULL) {
		device_printf(dev, "couldn't allocate scratch buffer (%d bytes)\n", scratch_size);
		goto config_failed;
	}

	/* 
	 * Allocate RX/TX Pools
	 */
	for (i = 0; i < RING_BUFFER_LEN; i++) {
		sc->rx_ring[i].index = i;
		sc->rx_ring[i].data = (char *)malloc(RX_BUFFER_LEN, M_DEVBUF, M_NOWAIT); 
		sc->rx_ring[i].address = vtophys(sc->rx_ring[i].data);
		sc->tx_ring[i].index = i;
		sc->tx_ring[i].data = (char *)malloc(TX_BUFFER_LEN, M_DEVBUF, M_NOWAIT); 
		sc->tx_ring[i].address = vtophys(sc->tx_ring[i].data);
		if ((!sc->rx_ring[i].data) || (!sc->tx_ring[i].data)) {
			device_printf(dev, "unable to allocate ring buffers\n");
			while (i > 0) {
				if (sc->rx_ring[i].data)
					free(sc->rx_ring[i].data, M_DEVBUF);
				if (sc->tx_ring[i].data)
					free(sc->tx_ring[i].data, M_DEVBUF);
				i--;
			}
			goto config_failed;
		}
	}
	
	/* 
	 * Initialize Adapter
	 */
	if ((rc = TRlldAdapterInit(&LldDriver, sc->TRlldAdapter, vtophys(sc->TRlldAdapter), 
                                   (void *)sc, &config)) != TRLLD_INIT_OK) {
		switch(rc) {
			case TRLLD_INIT_NOT_FOUND:
				device_printf(dev, "adapter not found\n");
				break;
			case TRLLD_INIT_UNSUPPORTED:
				device_printf(dev, "adapter not supported by low level driver\n");
				break;
			case TRLLD_INIT_PHYS16:
				device_printf(dev, "adapter memory block above 16M cannot DMA\n");
				break;
			case TRLLD_INIT_VERSION:
				device_printf(dev, "low level driver version mismatch\n");
				break;
			default:
				device_printf(dev, "unknown init error %d\n", rc);
				break;
			goto config_failed;
		}
	}
	sc->state = OL_INIT;

	/* 
	 * Download adapter micro-code
	 */
	if (bootverbose)
		device_printf(dev, "Downloading adapter microcode: ");

	switch(config.mactype) {
		case TRLLD_MAC_TMS:
			rc = TRlldDownload(sc->TRlldAdapter, TRlldMacCode);
			if (bootverbose)
				printf("TMS-380");
			break;
		case TRLLD_MAC_HAWKEYE:
			rc = TRlldDownload(sc->TRlldAdapter, TRlldHawkeyeMac);
			if (bootverbose)
				printf("Hawkeye");
			break;
		case TRLLD_MAC_BULLSEYE:
			rc = TRlldDownload(sc->TRlldAdapter, TRlldBullseyeMac);
			if (bootverbose)
				printf("Bullseye");
			break;
		default:
			if (bootverbose)
				printf("unknown - failed!\n");
			goto config_failed;
			break;
	}

	/* 
	 * Check download status
	 */
	switch(rc) {
		case TRLLD_DOWNLOAD_OK:
			if (bootverbose)
				printf(" - ok\n");
			break;
		case TRLLD_DOWNLOAD_ERROR:
			if (bootverbose)
				printf(" - failed\n");
			else 
				device_printf(dev, "adapter microcode download failed\n");
			goto config_failed;
			break;
		case TRLLD_STATE:
			if (bootverbose)
				printf(" - not ready\n");
			goto config_failed;
			break;
	}

	/* 
	 * Allocate interrupt and DMA channel
	 */
	rid = 0;
	sc->ol_irq = bus_alloc_resource(dev, SYS_RES_IRQ, &rid, 0, ~0, 1,
		(config.mode & TRLLD_MODE_SHARE_INTERRUPT ? RF_ACTIVE | RF_SHAREABLE : RF_ACTIVE));
	if (sc->ol_irq == NULL) {
		device_printf(dev, "couldn't map interrupt\n");
		goto config_failed;
	}
	if (bus_setup_intr(dev, sc->ol_irq, INTR_TYPE_NET, ol_intr, sc, &sc->ol_intrhand)) {
		device_printf(dev, "couldn't setup interrupt\n");
		bus_release_resource(dev, SYS_RES_IRQ, 0, sc->ol_irq);
		goto config_failed;
	}

	/*
	 * Wait for self-test to complete 
	 */
	i = 0;
	while ((poll++ < SELF_TEST_POLLS) && (sc->state < OL_READY))
		DELAY(TRlldPoll(sc->TRlldAdapter) * 1000);

	if (sc->state != OL_CLOSED) {
		device_printf(dev, "self-test failed\n");
		goto init_failed;
	}

	/* 
	 * Set up adapter poll
	 */
	callout_handle_init(&sc->ol_poll_ch);
	sc->ol_poll_ch = timeout(ol_poll, (void *)sc, 1);

	/* 
	 * Do the ifnet initialization
	 */
	ifp->if_softc	= sc;
	ifp->if_unit	= device_get_unit(dev);
	ifp->if_name	= "ol";
	ifp->if_output	= iso88025_output;
	ifp->if_init	= ol_init;
	ifp->if_start	= ol_start;
	ifp->if_ioctl	= ol_ioctl;
	ifp->if_flags	= IFF_BROADCAST | IFF_MULTICAST;
	bcopy(config.macaddress, sc->arpcom.ac_enaddr, sizeof(config.macaddress));

	/* 
	 * Attach the interface
	 */
	if_attach(ifp);
	ifp->if_snd.ifq_maxlen = IFQ_MAXLEN;
	iso88025_ifattach(ifp);

#if NBPF > 0
	bpfattach(ifp, DLT_IEEE802, sizeof(struct iso88025_header));
#endif

	splx(s);
	return(0);

init_failed:
	bus_teardown_intr(dev, sc->ol_irq, sc->ol_intrhand);
	bus_release_resource(dev, SYS_RES_IRQ, 0, sc->ol_irq);
config_failed:
        (void)splx(s); 
       
	return(ENXIO);
}

static int ol_pci_detach(dev)
	device_t	dev;
{
	struct ol_softc		*sc = device_get_softc(dev);
	struct ifnet		*ifp = &sc->arpcom.ac_if;
	int s;

	device_printf(dev, "driver unloading\n");

	s = splimp();

	if_detach(ifp);
	ol_stop(sc);

	untimeout(ol_poll, (void *)sc, sc->ol_poll_ch);
	/*untimeout(ol_stat, (void *)sc, sc->ol_stat_ch);*/

	bus_teardown_intr(dev, sc->ol_irq, sc->ol_intrhand);
	bus_release_resource(dev, SYS_RES_IRQ, 0, sc->ol_irq);

	(void)splx(s);

	return(0);
}

static void ol_pci_shutdown(dev)
	device_t		dev;  
{
	struct ol_softc		*sc = device_get_softc(dev);

	device_printf(dev, "ol_pci_shutdown called\n");

	ol_stop(sc);

	return;
}

static void ol_intr(xsc)
	void			*xsc;
{
	struct ol_softc		*sc = (struct ol_softc *)xsc;

	if (DEBUG_MASK & DEBUG_INT) printf("I");

	if (TRlldInterruptService(sc->TRlldAdapter) == 0)
		printf("ol%d: spurious interrupt\n", sc->unit);

	return;
}

static void ol_start(ifp)
	struct ifnet		*ifp;
{
	struct ol_softc 	*sc = ifp->if_softc;
	struct mbuf		*m0, *m;
	int			copy_len, buffer, frame, fragment, rc, s;
	
	/* 
	 * Check to see if output is already active
	 */
	if (ifp->if_flags & IFF_OACTIVE)
		return;

outloop:

	/*
	 * Make sure we have buffers to transmit with
	 */
	if (sc->tx_avail <= 0) {
		printf("ol%d: tx queue full\n", sc->unit);
		ifp->if_flags |= IFF_OACTIVE;
		return;
	}

	if (sc->restart == NULL) {
		IF_DEQUEUE(&ifp->if_snd, m);
		if (m == NULL) 
			return;
	} else {
		m = sc->restart;
		sc->restart = NULL;
	}

	m0 = m;
	frame = RING_BUFFER(sc->tx_frame);
	buffer = RING_BUFFER(sc->tx_head);
	fragment = 0;
	copy_len = 0;
	sc->frame_ring[frame].FragmentCount = 0;
	
	while (copy_len < m0->m_pkthdr.len) {
		sc->frame_ring[frame].FragmentCount++;
		if (sc->frame_ring[frame].FragmentCount > sc->tx_avail)
			goto nobuffers;
		sc->frame_ring[frame].TransmitFragment[fragment].VirtualAddress = sc->tx_ring[buffer].data;
		sc->frame_ring[frame].TransmitFragment[fragment].PhysicalAddress = sc->tx_ring[buffer].address;
		sc->frame_ring[frame].TransmitFragment[fragment].count = MIN(m0->m_pkthdr.len - copy_len, TX_BUFFER_LEN);
		m_copydata(m0, copy_len, MIN(m0->m_pkthdr.len - copy_len, TX_BUFFER_LEN), sc->tx_ring[buffer].data);
		copy_len += MIN(m0->m_pkthdr.len - copy_len, TX_BUFFER_LEN);
		fragment++;
		buffer = RING_BUFFER((buffer + 1));
	}

	s = splimp();
	rc = TRlldTransmitFrame(sc->TRlldAdapter, &sc->frame_ring[frame], (void *)&sc->frame_ring[frame]);
	(void)splx(s);

	if (rc != TRLLD_TRANSMIT_OK) {
		printf("ol%d: TRlldTransmitFrame returned %d\n", sc->unit, rc);
		ifp->if_oerrors++;
		goto bad;
	}

	sc->tx_avail -= sc->frame_ring[frame].FragmentCount++;
	sc->tx_head += sc->frame_ring[frame].FragmentCount++;
	sc->tx_frame++;

bad:
	m_freem(m0);

	goto outloop;

nobuffers:

	printf("ol%d: queue full\n", sc->unit);
	ifp->if_flags |= IFF_OACTIVE;
	ifp->if_oerrors++;
	/*m_freem(m0);*/
	sc->restart = m0;

	return;
}

static void ol_stop(sc)
	struct ol_softc		*sc;
{
	struct ifnet 		*ifp = &sc->arpcom.ac_if;

	printf("ol%d: ol_stop\n", sc->unit);

	ifp->if_flags &= ~(IFF_UP | IFF_RUNNING | IFF_OACTIVE);
	TRlldClose(sc->TRlldAdapter, 0);
	sc->state = OL_CLOSING;
}

static void ol_init(xsc)
	void *			xsc;
{
	struct ol_softc 	*sc = (struct ol_softc *)xsc;
	struct ifnet		*ifp = &sc->arpcom.ac_if;
	int			i, rc, s;

	/*
	 * Check adapter state, don't allow multiple inits
	 */
	if (sc->state > OL_CLOSED) {
		printf("ol%d: adapter not ready\n", sc->unit);
		return;
	}
	sc->state = OL_OPENING;

	s = splimp();

	/*
	 * Open the adapter
	 */
	rc = TRlldOpen(sc->TRlldAdapter, sc->arpcom.ac_enaddr, sc->GroupAddress,
		sc->FunctionalAddress, 1552, sc->AdapterMode);
	switch(rc) {
		case TRLLD_OPEN_OK:
			break;
		case TRLLD_OPEN_STATE:
			printf("ol%d: adapter not ready for open\n", sc->unit);
			(void)splx(s);
			return;
		case TRLLD_OPEN_ADDRESS_ERROR:
			printf("ol%d: illegal MAC address\n", sc->unit);
			(void)splx(s);
			return;
		case TRLLD_OPEN_MODE_ERROR:
			printf("ol%d: illegal open mode\n", sc->unit);
			(void)splx(s);
			return;
		default:
			printf("ol%d: unkown open error (%d)\n", sc->unit, rc);
			(void)splx(s);
			return;
	}

	/*
	 * Set promiscious mode for now...
	 */
	TRlldSetPromiscuousMode(sc->TRlldAdapter, TRLLD_PROM_LLC);

	/*
	 * Block on the ring insert and set a timeout
	 */

	/*
	 * Set up receive buffer ring 
	 */

	for (i = 0; i < RING_BUFFER_LEN; i++) {
		rc = TRlldReceiveFragment(sc->TRlldAdapter, (void *)sc->rx_ring[i].data,
			sc->rx_ring[i].address, RX_BUFFER_LEN, (void *)sc->rx_ring[i].index);
		if (rc != TRLLD_RECEIVE_OK) {
			printf("ol%d: adapter refused receive fragment %d (rc = %d)\n", sc->unit, i, rc);
			break;
		}	
	}

	sc->tx_avail = RING_BUFFER_LEN;
	sc->tx_head = 0;
	sc->tx_frame = 0;

	sc->restart = NULL;

	ifp->if_flags |= IFF_RUNNING;
	ifp->if_flags &= ~IFF_OACTIVE;

	/* 
	 * Set up adapter statistics poll
	 */
	/*callout_handle_init(&sc->ol_stat_ch);*/
	/*sc->ol_stat_ch = timeout(ol_stat, (void *)sc, 1*hz);*/


	(void)splx(s);

}

static int ol_ioctl(ifp, command, data)
	struct ifnet 	*ifp;
	u_long 		command;
	caddr_t 	data;
{
	struct ol_softc 	*sc = ifp->if_softc;
	/*struct ifreq		*ifr = (struct ifreq *)data;*/
	int 			error = 0, s;

	s = splimp();

	switch(command) {
		case SIOCSIFADDR:
		case SIOCGIFADDR:
		case SIOCSIFMTU:
			error = iso88025_ioctl(ifp, command, data);
			break;

		case SIOCSIFFLAGS:
			if (ifp->if_flags & IFF_UP) {
				ol_init(sc);
			} else {
				if (ifp->if_flags & IFF_RUNNING)
					ol_stop(sc);
			}
			break;
	}

	(void)splx(s);	

	return(error);
}


void ol_poll(arg)
	void *arg;
{
	struct ol_softc *sc = (struct ol_softc *)arg;
	int s;

	s = splimp();

	if (DEBUG_MASK & DEBUG_POLL) printf("P");

	/* Set up next adapter poll */
	sc->ol_poll_ch = timeout(ol_poll, (void *)sc, (TRlldPoll(sc->TRlldAdapter) * hz / 1000));

	(void)splx(s);
}

#ifdef NOTYET
void ol_stat(arg)
	void *arg;
{
	struct ol_softc		*sc = (struct ol_softc *)arg;
	int			s;

	s = splimp();

	/* Set up next adapter poll */
	sc->ol_stat_ch = timeout(ol_stat, (void *)sc, 1*hz);
	if (TRlldGetStatistics(sc->TRlldAdapter, &sc->current, 0) != 0) {
		/*printf("ol%d: statistics available immediately...\n", sc->unit);*/
		DriverStatistics((void *)sc, &sc->current);
	}

	(void)splx(s);
}
#endif

/*
 * ---------------------- PMW Callback Functions -----------------------
 */

void DriverStatistics(DriverHandle, statistics)
	void			*DriverHandle;
	TRlldStatistics_t	*statistics;
{
#ifdef NOTYET
	struct ol_softc		*sc = (struct ol_softc *)DriverHandle;

	if (sc->statistics.LineErrors != statistics->LineErrors)
		printf("ol%d: Line Errors %lu\n", sc->unit, statistics->LineErrors);
	if (sc->statistics.InternalErrors != statistics->InternalErrors)
		printf("ol%d: Internal Errors %lu\n", sc->unit, statistics->InternalErrors);
	if (sc->statistics.BurstErrors != statistics->BurstErrors)
		printf("ol%d: Burst Errors %lu\n", sc->unit, statistics->BurstErrors);
	if (sc->statistics.AbortDelimiters != statistics->AbortDelimiters)
		printf("ol%d: Abort Delimiters %lu\n", sc->unit, statistics->AbortDelimiters);
	if (sc->statistics.ARIFCIErrors != statistics->ARIFCIErrors)
		printf("ol%d: ARIFCI Errors %lu\n", sc->unit, statistics->ARIFCIErrors);
	if (sc->statistics.LostFrames != statistics->LostFrames)
		printf("ol%d: Lost Frames %lu\n", sc->unit, statistics->LostFrames);
	if (sc->statistics.CongestionErrors != statistics->CongestionErrors)
		printf("ol%d: Congestion Errors %lu\n", sc->unit, statistics->CongestionErrors);
	if (sc->statistics.FrequencyErrors != statistics->FrequencyErrors)
		printf("ol%d: Frequency Errors %lu\n", sc->unit, statistics->FrequencyErrors);
	if (sc->statistics.TokenErrors != statistics->TokenErrors)
		printf("ol%d: Token Errors %lu\n", sc->unit, statistics->TokenErrors);
	if (sc->statistics.DMABusErrors != statistics->DMABusErrors)
		printf("ol%d: DMA Bus Errors %lu\n", sc->unit, statistics->DMABusErrors);
	if (sc->statistics.DMAParityErrors != statistics->DMAParityErrors)
		printf("ol%d: DMA Parity Errors %lu\n", sc->unit, statistics->DMAParityErrors);
	if (sc->statistics.ReceiveLongFrame != statistics->ReceiveLongFrame)
		printf("ol%d: Long frames received %lu\n", sc->unit, statistics->ReceiveLongFrame);
	if (sc->statistics.ReceiveCRCErrors != statistics->ReceiveCRCErrors)
		printf("ol%d: Receive CRC Errors %lu\n", sc->unit, statistics->ReceiveCRCErrors);
	if (sc->statistics.ReceiveOverflow != statistics->ReceiveOverflow)
		printf("ol%d: Recieve overflows %lu\n", sc->unit, statistics->ReceiveOverflow);
	if (sc->statistics.TransmitUnderrun != statistics->TransmitUnderrun)
		printf("ol%d: Frequency Errors %lu\n", sc->unit, statistics->TransmitUnderrun);
	bcopy(statistics, &sc->statistics, sizeof(TRlldStatistics_t));
#endif
}

static void
DriverSuspend(MicroSeconds)
	unsigned short		MicroSeconds;
{
    DELAY(MicroSeconds);
}


static void
DriverStatus(DriverHandle, Status)
	void			*DriverHandle;
	TRlldStatus_t		*Status;
{
	struct ol_softc		*sc = (struct ol_softc *)DriverHandle;
	
	switch (Status->Type) {

		case TRLLD_STS_ON_WIRE:
			printf("ol%d: ring insert\n", sc->unit);
			sc->state = OL_OPEN;
			break;
		case TRLLD_STS_SELFTEST_STATUS:
			if (sc->state == OL_INIT)
				printf("ol%d: adapter self-test complete (status=%d)\n", sc->unit, Status->Specification.SelftestStatus);
			else
				printf("ol%d: adapter closed\n", sc->unit);
			sc->state = OL_CLOSED;
			break;
		case TRLLD_STS_INIT_STATUS:
			printf("ol%d: adapter init failure %03x\n", sc->unit, Status->Specification.InitStatus);
			ol_stop(sc);
			break;
		case TRLLD_STS_ADAPTER_CHECK:
			printf("ol%d: adapter check (%04x %04x %04x %04x)\n", sc->unit, 
			    Status->Specification.AdapterCheck[0], 
			    Status->Specification.AdapterCheck[1], 
			    Status->Specification.AdapterCheck[2], 
			    Status->Specification.AdapterCheck[3]);
			sc->state = OL_UNKNOWN;
			ol_stop(sc);
			break;
		case TRLLD_STS_ADAPTER_TIMEOUT:
			switch(Status->Specification.AdapterTimeout) {
				case TRLLD_COMMAND_TIMEOUT:
					printf("ol%d: adapter command timeout\n", sc->unit);
					break;
				case TRLLD_TRANSMIT_TIMEOUT:
					printf("ol%d: adapter transmit timeout\n", sc->unit);
					break;
				case TRLLD_INTERRUPT_TIMEOUT:
					printf("ol%d: adapter interrupt timeout\n", sc->unit);
					break;
				default:
					printf("ol%d: unknown adapter timout %d\n", sc->unit, Status->Specification.AdapterTimeout);
			}
			break;
		default:
			printf("ol%d: driver status Type = %d\n", sc->unit, Status->Type);
			break;

	}
	if (Status->Closed) {
		sc->state = OL_CLOSING;
		ol_stop(sc);
	}

}

static void
DriverCloseCompleted(DriverHandle)
	void			*DriverHandle;
{
	struct ol_softc		*sc = (struct ol_softc *)DriverHandle;
	
	printf("ol%d: adapter closed\n", sc->unit);
	sc->state = OL_CLOSED;
}

static void
DriverTransmitFrameCompleted(DriverHandle, FrameHandle, TransmitStatus)
	void			*DriverHandle;
	void			*FrameHandle;
	int			TransmitStatus;
{
	struct ol_softc		*sc = (struct ol_softc *)DriverHandle;
	struct ifnet		*ifp = &sc->arpcom.ac_if;
	TRlldTransmit_t		*frame = (TRlldTransmit_t *)FrameHandle;
	
	/*printf("ol%d: DriverTransmitFrameCompleted\n", sc->unit);*/

	if (TransmitStatus == TRLLD_TRANSMIT_OK) {
		ifp->if_opackets++;
	} else {
		ifp->if_oerrors++;
		printf("ol%d: transmit error %d\n", sc->unit, TransmitStatus);
	}
	
	sc->tx_avail += frame->FragmentCount;

	if (ifp->if_flags & IFF_OACTIVE) {
		printf("ol%d: queue restart\n", sc->unit);
		ifp->if_flags &= ~IFF_OACTIVE;
		ol_start(ifp);
	}


}

static void
DriverReceiveFrameCompleted(DriverHandle, ByteCount, FragmentCount, FragmentHandle, ReceiveStatus)
	void 			*DriverHandle;
	int 			ByteCount;
	int 			FragmentCount;
	void 			*FragmentHandle;
	int 			ReceiveStatus;
{
	struct ol_softc 	*sc = (struct ol_softc *)DriverHandle;
	struct ifnet		*ifp = (struct ifnet *)&sc->arpcom.ac_if;
	struct mbuf		*m0, *m1, *m;
	struct iso88025_header	*th;
	int			frame_len = ByteCount, hdr_len, i = (int)FragmentHandle, rc, s;
	int			mbuf_offset, mbuf_size, frag_offset, copy_length;
	char			*fragment = sc->rx_ring[RING_BUFFER(i)].data;
	
	if (sc->state > OL_CLOSED) {
		if (ReceiveStatus == TRLLD_RCV_OK) {
			MGETHDR(m0, M_DONTWAIT, MT_DATA);
			mbuf_size = MHLEN - 2;
			if (!m0) {
				ifp->if_ierrors++;
				goto dropped;
			}
			if (ByteCount + 2 > MHLEN) {
				MCLGET(m0, M_DONTWAIT);
				mbuf_size = MCLBYTES - 2;
				if (!(m0->m_flags & M_EXT)) {
					m_freem(m0);
					ifp->if_ierrors++;
					goto dropped;
				}
			}
			m0->m_pkthdr.rcvif = ifp;
			m0->m_pkthdr.len = ByteCount;
			m0->m_len = 0;
			m0->m_data += 2;
			th = mtod(m0, struct iso88025_header *);
			m0->m_pkthdr.header = (void *)th;

			m = m0;
			mbuf_offset = 0;
			frag_offset = 0;
			while (frame_len) {
				copy_length = MIN3(frame_len, (RX_BUFFER_LEN - frag_offset), 
						(mbuf_size - mbuf_offset));
				bcopy(fragment + frag_offset, mtod(m, char *) + mbuf_offset, copy_length);
				m->m_len += copy_length;
				mbuf_offset += copy_length;
				frag_offset += copy_length;
				frame_len -= copy_length;
			
				if (frag_offset == RX_BUFFER_LEN) {
					fragment = sc->rx_ring[RING_BUFFER(++i)].data;
					frag_offset = 0;
				}
				if ((mbuf_offset == mbuf_size) && (frame_len > 0)) {
					MGET(m1, M_DONTWAIT, MT_DATA);
					mbuf_size = MHLEN;
					if (!m1) {
						ifp->if_ierrors++;
						m_freem(m0);
						goto dropped;
					}
					if (frame_len > MHLEN) {
						MCLGET(m1, M_DONTWAIT);
						mbuf_size = MCLBYTES;
						if (!(m1->m_flags & M_EXT)) {
							m_freem(m0);
							m_freem(m1);
							ifp->if_ierrors++;
							goto dropped;
						}
					}
					m->m_next = m1;
					m = m1;
					mbuf_offset = 0;
					m->m_len = 0;
				}
			}
			ifp->if_ipackets++;
#if NBPF
			if (ifp->if_bpf)
				bpf_mtap(ifp, m0);
#endif

			/*if (ifp->if_flags & IFF_PROMISC) {*/
				if (bcmp(th->iso88025_dhost, etherbroadcastaddr , sizeof(th->iso88025_dhost))) {
					if ((bcmp(th->iso88025_dhost + 1, sc->arpcom.ac_enaddr + 1, ISO88025_ADDR_LEN - 1)) || 
					    ((th->iso88025_dhost[0] & 0x7f) != sc->arpcom.ac_enaddr[0])) {
						m_freem(m0);
						goto dropped;
					}
				}
			/*}*/

			hdr_len = ISO88025_HDR_LEN;
			if (th->iso88025_shost[0] & 0x80)	/* Check for source route */
				hdr_len += (ntohs(th->rcf) & 0x1f00) >> 8;

			m0->m_pkthdr.len -= hdr_len;
			m0->m_len -= hdr_len;
			m0->m_data += hdr_len;

			iso88025_input(ifp, th, m0);

		} else {	/* Receiver error */
			if (ReceiveStatus != TRLLD_RCV_NO_DATA) {
				printf("ol%d: receive error %d\n", sc->unit, ReceiveStatus);
				ifp->if_ierrors++;
			}
		}

dropped:
		s = splimp();
		i = (int)FragmentHandle;
		while (FragmentCount--) {
			rc = TRlldReceiveFragment(sc->TRlldAdapter, (void *)sc->rx_ring[RING_BUFFER(i)].data, 
				sc->rx_ring[RING_BUFFER(i)].address, RX_BUFFER_LEN, (void *)sc->rx_ring[RING_BUFFER(i)].index);
			if (rc == TRLLD_RECEIVE_OK) {
			} else {
				printf("ol%d: adapter refused receive fragment %d (rc = %d)\n", sc->unit, i, rc);
				break;
			}
			i++;
		}
		(void)splx(s);
	}
}


/*
 * ---------------------------- PMW Glue -------------------------------
 */

#ifndef TRlldInlineIO

static void 
DriverOutByte(IOAddress, value)
	unsigned short	IOAddress;
	unsigned char	value;
{
	outb(IOAddress, value);
}

static void
DriverOutWord(IOAddress, value)
	unsigned short	IOAddress;
	unsigned short	value;
{
	outw(IOAddress, value);
}

static void
DriverOutDword(IOAddress, value)
	unsigned short	IOAddress;
	unsigned long	value;
{
	outl(IOAddress, value);
}

static void
DriverRepOutByte(IOAddress, DataPointer, ByteCount)
	unsigned short	IOAddress;
	unsigned char	*DataPointer;
	int 		ByteCount;
{
	outsb(IOAddress, (void *)DataPointer, ByteCount);
}

static void
DriverRepOutWord(IOAddress, DataPointer, WordCount)
	unsigned short	IOAddress;
	unsigned short	*DataPointer;
	int 		WordCount;
{
	outsw(IOAddress, (void *)DataPointer, WordCount);
}

static void
DriverRepOutDword(IOAddress, DataPointer, DWordCount)
	unsigned short	IOAddress;
	unsigned long	*DataPointer;
	int 		DWordCount;
{
	outsl(IOAddress, (void *)DataPointer, DWordCount);
}

static unsigned char
DriverInByte(IOAddress)
	unsigned short IOAddress;
{
	return(inb(IOAddress));
}

static unsigned short
DriverInWord(IOAddress)
	unsigned short IOAddress;
{
	return(inw(IOAddress));
}

static unsigned long
DriverInDword(IOAddress)
	unsigned short IOAddress;
{
	return(inl(IOAddress));
}

static void
DriverRepInByte(IOAddress, DataPointer, ByteCount)
	unsigned short	IOAddress;
	unsigned char	*DataPointer;
	int 		ByteCount;
{
	insb(IOAddress, (void *)DataPointer, ByteCount);
}

static void
DriverRepInWord(IOAddress, DataPointer, WordCount)
	unsigned short	IOAddress;
	unsigned short	*DataPointer;
	int 		WordCount;
{
	insw(IOAddress, (void *)DataPointer, WordCount);
}
static void
DriverRepInDword(IOAddress, DataPointer, DWordCount)
	unsigned short	IOAddress;
	unsigned long	*DataPointer;
	int 		DWordCount;
{
	insl(IOAddress, (void *)DataPointer, DWordCount);
}
#endif /* TRlldInlineIO */
